import gc
import numpy as np


def generate_new_pick(initial_stripe, overlapping_stripes):
    new_stripe = np.squeeze(initial_stripe) + 0
    new_stripe[2] = min(new_stripe[2], np.min(overlapping_stripes[:, 2]))
    new_stripe[3] = max(new_stripe[3], np.max(overlapping_stripes[:, 3]))
    return new_stripe


def get_overlapping_indices(pick, indexes, last, boxes):
    xx1 = np.maximum(pick[0], boxes[indexes[:last], 0])
    xx2 = np.minimum(pick[1], boxes[indexes[:last], 1])
    yy1 = np.maximum(pick[2], boxes[indexes[:last], 2])
    yy2 = np.minimum(pick[3], boxes[indexes[:last], 3])
    area = np.maximum(0, xx2 - xx1) * np.maximum(0, yy2 - yy1)
    indices = np.where(area > 0)[0]
    return indices


def update_stripe_merger(pick, indexes, last, boxes):
    overlapping_indices = get_overlapping_indices(pick, indexes, last, boxes)
    if len(overlapping_indices) > 0:
        return generate_new_pick(pick, boxes[indexes[overlapping_indices]]), overlapping_indices
    return pick, overlapping_indices


def nms(boxes, assimilate: bool = False):
    if len(boxes) == 0:
        return []
    picks = []
    indexes = np.argsort(boxes[:, 4])
    num_boxes_left = len(indexes)
    while num_boxes_left > 0:
        last = num_boxes_left - 1
        i = indexes[last]
        pick = np.squeeze(boxes[i].copy())
        num_overlap = 0
        if assimilate:
            while True:
                pick, overlapping_indices = update_stripe_merger(pick, indexes, last, boxes)
                if num_overlap == len(overlapping_indices):
                    break
                num_overlap = len(overlapping_indices)
        else:
            overlapping_indices = get_overlapping_indices(boxes[i], indexes, last, boxes)
        overlapping_indices = np.concatenate(([last], overlapping_indices))
        indexes = np.delete(indexes, overlapping_indices)
        picks.append(pick)
        num_boxes_left = len(indexes)
    return np.asarray(picks)


class Handler:
    def __init__(self):
        self.findings = {}

    def do_nms_and_print_to_file(self, outfile_name, is_domain=False, is_stripe=False):
        for key in self.findings:
            temp = list(self.findings[key])
            temp = np.vstack(temp)
            if is_stripe:
                self.findings[key] = nms(temp, assimilate=True)
            else:
                self.findings[key] = nms(temp)
        outfile_handler = open(outfile_name, 'w')
        outfile_handler.write("#chr1\tx1\tx2\tchr2\ty1\ty2\tname1\tscore\tstrand1\tstrand2\tcolor\n")
        for key in self.findings:
            self.write_to_file(outfile_handler, key[0], key[1], self.findings[key], is_domain=is_domain)
        outfile_handler.close()

    def add_prediction(self, chrom1, chrom2, x1, x2, y1, y2, prediction, offset=0, priority=0):
        local_key = (chrom1, chrom2)
        if local_key not in self.findings:
            self.findings[local_key] = set()
        self.findings[local_key].add((x1, x2, y1, y2, prediction, offset, priority))

    @staticmethod
    def parse_bedpe_line(data, index):
        x1, x2 = int(data[index, 0]), int(data[index, 1])
        y1, y2, = int(data[index, 2]), int(data[index, 3])
        score = float(data[index, 4])
        return x1, x2, y1, y2, score

    def write_to_file(self, outfile_handler, chrom1, chrom2, results, is_domain=False):
        for k in range(np.shape(results)[0]):
            x1, x2, y1, y2, score = self.parse_bedpe_line(results, k)
            if is_domain:
                self.write_line(outfile_handler, chrom1, x1, y2, chrom1, x1, y2, '0,0,0', score)
            else:
                self.write_line(outfile_handler, chrom1, x1, x2, chrom2, y1, y2, '0,0,0', score)

    @staticmethod
    def write_line(out_file_handler, chrom1, x1, x2, chrom2, y1, y2, color, prediction):
        out_file_handler.write("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t.\t{6}\t.\t.\t{7}\n".
                               format(chrom1, x1, x2, chrom2, y1, y2, prediction, color))


class MultiResHandler(Handler):
    def __init__(self, files, output_name, radii, is_domain=False, is_stripe=False):
        super().__init__()
        for f in range(len(files)):
            self.fill_out_all_indices(files[f], len(files) - f, radii[f], is_stripe=is_stripe)
        self.do_nms_and_print_to_file(output_name, is_domain=is_domain, is_stripe=is_stripe)

    def fill_out_all_indices(self, infile, priority, offset, is_stripe):
        with open(infile) as f:
            for line in f:
                if line.startswith('#'):
                    continue
                line_split = line.split()
                chrom1, x1, x2 = line_split[0], int(line_split[1]) - offset, int(line_split[2]) + offset
                chrom2, y1, y2 = line_split[3], int(line_split[4]) - offset, int(line_split[5]) + offset
                prediction = float(line_split[7]) + priority
                if is_stripe and x2 - x1 > y2 - y1:
                    self.add_prediction(chrom1, chrom2, y1, y2, x1, x2, prediction, offset, priority)
                else:
                    self.add_prediction(chrom1, chrom2, x1, x2, y1, y2, prediction, offset, priority)
        gc.collect()

    def write_to_file(self, outfile_handler, chrom1, chrom2, results, is_domain=False):
        for k in range(np.shape(results)[0]):
            x1, x2, y1, y2, score = self.parse_bedpe_line(results, k)
            offset, priority = int(results[k, 5]), int(results[k, 6])
            if is_domain:
                self.write_line(outfile_handler, chrom1, x1 + offset, y2 - offset, chrom1, x1 + offset, y2 - offset,
                                '0,0,0', score - priority)
            else:
                self.write_line(outfile_handler, chrom1, x1 + offset, x2 - offset, chrom2, y1 + offset, y2 - offset,
                                '0,0,0', score - priority)



from keras import models, layers
from keras import backend as keras
import tensorflow as tf
import time
import glob
import numpy as np

from hicml.StrawMLTools import DeploySpears
from hicml.NonMaxSuppression import MultiResHandler


class WBCE:
    def __init__(self, weight: int = 120):
        self.weight = weight

    def func(self, y_true, y_pred):
        y_true = keras.clip(y_true, keras.epsilon(), 1 - keras.epsilon())
        y_pred = keras.clip(y_pred, keras.epsilon(), 1 - keras.epsilon())
        return -keras.mean(y_true[:, :, :, 0] * keras.log(y_pred[:, :, :, 0]) * self.weight +
                           (1 - y_true[:, :, :, 0]) * keras.log(1 - y_pred[:, :, :, 0]))


def preprocessing_method(matrix, m_width: int = 500):  # , scale=0.1
    result = np.zeros((m_width, m_width, 3))
    if np.sum(matrix) < 1e-9:
        return result
    flattened_data = matrix.flatten()
    flattened_data = flattened_data[flattened_data > 0]
    median_val = np.median(flattened_data)
    mad_val = np.median(np.abs(flattened_data - median_val))
    if mad_val < 1:
        mad_val = 1
    m2 = matrix / mad_val
    result[:, :, 0] = m2 / 5
    result[:, :, 1] = m2 / 10
    result[:, :, 2] = m2 / 30
    return np.tanh(result - 2)


def load_models(feature_types: list, model_directory: str) -> list:
    all_models = []
    wbce = WBCE()
    print("Loading all Models")
    for feature in feature_types:
        model_list = []
        for file in glob.glob(model_directory + "/*" + feature + "*.h5"):
            print("Using Model", file, flush=True)
            model_list.append(models.load_model(file,
                                                custom_objects={'wbce': wbce.func,
                                                                'func': wbce.func,
                                                                'leaky_relu': tf.nn.leaky_relu,
                                                                'LeakyReLU': layers.LeakyReLU()}))
        all_models.append(model_list)
    return all_models


def deploy_at_res(filepath: str, all_models: list, matrix_width: int, res: int, working_directory: str,
                  normalization_type: str, stem: str, max_examples_in_ram: int, num_straw_workers: int,
                  threshold: float, batch_size: int, chromosome_subset: list = None):
    out_path = working_directory + "/" + stem
    loop_file = out_path + '_loops_' + str(res) + '.bedpe'
    domain_file = out_path + '_domains_' + str(res) + '.bedpe'
    stripe_file = out_path + '_stripes_' + str(res) + '.bedpe'
    loop_domain_file = out_path + '_loop_domains_' + str(res) + '.bedpe'
    print('got to deployment')
    print("Doing resolution", res, "for", stem, flush=True)
    DeploySpears(all_model_sets=all_models, batch_size=batch_size,
                 num_straw_workers=num_straw_workers, filepath=filepath,
                 resolution=res,
                 max_examples_in_ram=max_examples_in_ram, matrix_width=matrix_width,
                 threshold=threshold,
                 out_files=[loop_file, domain_file, stripe_file, loop_domain_file],
                 preprocess_method=preprocessing_method,
                 use_arithmetic_mean=False,
                 norm=normalization_type,
                 num_output_channels=3, chromosome_subset=chromosome_subset)
    return loop_file, domain_file, stripe_file, loop_domain_file


def merge_lists(working_directory, stem, loop_lists, domain_lists, stripe_lists,
                loop_domain_lists, loop_radii, domain_radii):
    if len(loop_lists) > 1:
        out_path = working_directory + "/" + stem
        MultiResHandler(loop_lists, out_path + '_loops_merged.bedpe', radii=loop_radii)
        MultiResHandler(domain_lists, out_path + '_domains_merged.bedpe', radii=domain_radii, is_domain=True)
        MultiResHandler(stripe_lists, out_path + '_stripes_merged.bedpe', radii=loop_radii, is_stripe=True)
        MultiResHandler(loop_domain_lists, out_path + '_loop_domains_merged.bedpe', radii=domain_radii, is_domain=True)


class DeployTridentFeatures:
    def __init__(self, filepath: str, all_models: list, resolution_strings: list,
                 stem: str, working_directory: str, normalization_type: str, threshold: float, matrix_width: int,
                 max_examples_in_ram: int, num_straw_workers: int, batch_size: int,
                 chromosome_subset: list = None):
        resolutions = []
        for res_string in resolution_strings:
            resolutions.append(int(res_string))
        resolutions.sort()

        loop_lists = []
        domain_lists = []
        stripe_lists = []
        loop_domains_lists = []
        loop_radii = []
        domain_radii = []
        for res in resolutions:
            start_time = time.time()
            loops, domains, stripes, loop_domains = deploy_at_res(filepath, all_models, matrix_width, res,
                                                                  working_directory, normalization_type, stem,
                                                                  max_examples_in_ram, num_straw_workers,
                                                                  threshold, batch_size, chromosome_subset)

            runtime = time.time() - start_time
            print("Executed resolution {} in {} seconds".format(res, runtime))
            print("-----------------------------------------", flush=True)

            loop_lists.append(loops)
            domain_lists.append(domains)
            stripe_lists.append(stripes)
            loop_domains_lists.append(loop_domains)
            loop_radii.append(res)
            domain_radii.append(3 * res)
        merge_lists(working_directory, stem, loop_lists, domain_lists, stripe_lists, loop_domains_lists,
                    loop_radii, domain_radii)
